from datetime import datetime, timedelta, timezone
from http import HTTPStatus
from types import SimpleNamespace
from unittest import TestCase
from unittest.mock import MagicMock, patch

from django_workflows.users.services import auth_flow


class AuthFlowTests(TestCase):
    @staticmethod
    def _make_user_model() -> SimpleNamespace:
        class DoesNotExist(Exception):
            pass

        return SimpleNamespace(DoesNotExist=DoesNotExist, objects=MagicMock())

    @staticmethod
    def _make_meta_model(meta: SimpleNamespace) -> SimpleNamespace:
        manager = MagicMock()
        manager.get_or_create.return_value = (meta, True)
        return SimpleNamespace(objects=manager)

    @staticmethod
    def _settings_getter(**overrides):
        defaults = {
            "COMPANY_NAME": "Acme Co",
            "ENABLE_EMAIL_DELIVERY": False,
            "RESET_CODE_TTL_MINUTES": 10,
            "MAX_VERIFY_ATTEMPTS": 3,
        }
        defaults.update(overrides)

        def _get(name, default=None):
            return defaults.get(name, default)

        return _get

    @patch.object(auth_flow, "get_user_model")
    def test_register_user_rejects_existing_email(self, mock_get_user_model):
        user_model = self._make_user_model()
        user_model.objects.filter.return_value.exists.return_value = True
        mock_get_user_model.return_value = user_model

        result = auth_flow.register_user_and_send_verification_email(
            email="existing@example.com",
            first_name="First",
            last_name="Last",
            password="pw",
        )

        self.assertFalse(result.success)
        self.assertEqual(result.message, "A user with this email already exists.")
        user_model.objects.create_user.assert_not_called()

    @patch.object(auth_flow, "get_user_model")
    def test_register_user_integrity_error_logs_exception(self, mock_get_user_model):
        user_model = self._make_user_model()
        user_model.objects.filter.return_value.exists.return_value = False
        user_model.objects.create_user.side_effect = auth_flow.IntegrityError("duplicate")
        mock_get_user_model.return_value = user_model

        with patch.object(auth_flow.logger, "exception") as mock_exception:
            result = auth_flow.register_user_and_send_verification_email(
                email="existing@example.com",
                first_name="First",
                last_name="Last",
                password="pw",
            )

        self.assertFalse(result.success)
        self.assertEqual(result.message, "Email already exists, login or reset password.")
        mock_exception.assert_called_once()

    @patch.object(auth_flow, "_register_on_commit", return_value=True)
    @patch.object(auth_flow, "timezone")
    @patch.object(auth_flow, "send_email")
    @patch.object(auth_flow, "generate_verification_code", return_value="ABC123")
    @patch.object(auth_flow, "_get_user_meta_model")
    @patch.object(auth_flow, "_get_workflow_setting")
    @patch.object(auth_flow, "get_user_model")
    def test_register_user_schedules_verification_email_on_commit(
        self,
        mock_get_user_model,
        mock_get_setting,
        mock_get_user_meta_model,
        _mock_generate_code,
        mock_send_email,
        mock_timezone,
        mock_register_on_commit,
    ):
        now = datetime(2026, 3, 14, 12, 0, tzinfo=timezone.utc)
        mock_timezone.now.return_value = now
        mock_send_email.return_value = SimpleNamespace(status_code=200, text="ok")
        mock_get_setting.side_effect = self._settings_getter(
            ENABLE_EMAIL_DELIVERY=True,
            COMPANY_NAME="Reusable Inc",
            RESET_CODE_TTL_MINUTES=15,
        )

        user_model = self._make_user_model()
        user_model.objects.filter.return_value.exists.return_value = False
        user = SimpleNamespace(email="new@example.com", first_name="New", last_name="User", delete=MagicMock())
        user_model.objects.create_user.return_value = user
        mock_get_user_model.return_value = user_model

        meta = SimpleNamespace(verify_code=None, verify_time=None, attempts=0, save=MagicMock())
        mock_get_user_meta_model.return_value = self._make_meta_model(meta)

        result = auth_flow.register_user_and_send_verification_email(
            email="new@example.com",
            first_name="New",
            last_name="User",
            password="pw",
        )

        self.assertTrue(result.success)
        mock_register_on_commit.assert_called_once()
        callback = mock_register_on_commit.call_args.args[0]
        self.assertTrue(callable(callback))
        mock_send_email.assert_not_called()
        callback()
        mock_send_email.assert_called_once()
        user.delete.assert_not_called()

    @patch.object(auth_flow, "timezone")
    @patch.object(auth_flow, "send_email")
    @patch.object(auth_flow, "generate_verification_code", return_value="ABC123")
    @patch.object(auth_flow, "_get_user_meta_model")
    @patch.object(auth_flow, "_get_workflow_setting")
    @patch.object(auth_flow, "get_user_model")
    def test_register_user_creates_meta_and_sends_email(
        self,
        mock_get_user_model,
        mock_get_setting,
        mock_get_user_meta_model,
        _mock_generate_code,
        mock_send_email,
        mock_timezone,
    ):
        now = datetime(2026, 3, 14, 12, 0, tzinfo=timezone.utc)
        mock_timezone.now.return_value = now
        mock_send_email.return_value = SimpleNamespace(status_code=200, text="ok")
        mock_get_setting.side_effect = self._settings_getter(
            ENABLE_EMAIL_DELIVERY=True,
            COMPANY_NAME="Reusable Inc",
            RESET_CODE_TTL_MINUTES=15,
        )

        user_model = self._make_user_model()
        user_model.objects.filter.return_value.exists.return_value = False
        user = SimpleNamespace(email="new@example.com", first_name="New", last_name="User", delete=MagicMock())
        user_model.objects.create_user.return_value = user
        mock_get_user_model.return_value = user_model

        meta = SimpleNamespace(verify_code=None, verify_time=None, attempts=0, save=MagicMock())
        mock_get_user_meta_model.return_value = self._make_meta_model(meta)

        result = auth_flow.register_user_and_send_verification_email(
            email="new@example.com",
            first_name="New",
            last_name="User",
            password="pw",
        )

        self.assertTrue(result.success)
        user_model.objects.create_user.assert_called_once()
        create_user_kwargs = user_model.objects.create_user.call_args.kwargs
        self.assertEqual(create_user_kwargs["email"], "new@example.com")
        self.assertEqual(create_user_kwargs["username"], "new@example.com")
        self.assertEqual(meta.verify_code, "ABC123")
        self.assertEqual(meta.verify_time, now)
        meta.save.assert_called_once()

        kwargs = mock_send_email.call_args.kwargs
        self.assertEqual(kwargs["to"], ["new@example.com"])
        self.assertEqual(kwargs["subject"], "Verify your account")
        self.assertTrue(kwargs["enabled"])
        self.assertIn("ABC123", kwargs["html"])
        self.assertIn("15 minutes", kwargs["html"])
        self.assertIn("Reusable Inc", kwargs["html"])
        user.delete.assert_not_called()

    @patch.object(auth_flow, "timezone")
    @patch.object(auth_flow, "send_email", return_value=None)
    @patch.object(auth_flow, "generate_verification_code", return_value="ABC123")
    @patch.object(auth_flow, "_get_user_meta_model")
    @patch.object(auth_flow, "_get_workflow_setting")
    @patch.object(auth_flow, "get_user_model")
    def test_register_user_returns_failure_when_verification_email_not_delivered(
        self,
        mock_get_user_model,
        mock_get_setting,
        mock_get_user_meta_model,
        _mock_generate_code,
        _mock_send_email,
        mock_timezone,
    ):
        now = datetime(2026, 3, 14, 12, 0, tzinfo=timezone.utc)
        mock_timezone.now.return_value = now
        mock_get_setting.side_effect = self._settings_getter(
            ENABLE_EMAIL_DELIVERY=True,
            COMPANY_NAME="Reusable Inc",
            RESET_CODE_TTL_MINUTES=15,
        )

        user_model = self._make_user_model()
        user_model.objects.filter.return_value.exists.return_value = False
        user = SimpleNamespace(email="new@example.com", first_name="New", last_name="User", delete=MagicMock())
        user_model.objects.create_user.return_value = user
        mock_get_user_model.return_value = user_model

        meta = SimpleNamespace(verify_code=None, verify_time=None, attempts=0, save=MagicMock())
        mock_get_user_meta_model.return_value = self._make_meta_model(meta)

        with patch.object(auth_flow.logger, "warning") as mock_warning:
            result = auth_flow.register_user_and_send_verification_email(
                email="new@example.com",
                first_name="New",
                last_name="User",
                password="pw",
            )

        self.assertFalse(result.success)
        self.assertIn("could not send a verification email", result.message.lower())
        mock_warning.assert_called_once()
        user.delete.assert_called_once()

    @patch.object(auth_flow, "get_user_model")
    def test_send_forgot_password_code_is_silent_for_missing_user(self, mock_get_user_model):
        user_model = self._make_user_model()
        user_model.objects.get.side_effect = user_model.DoesNotExist
        mock_get_user_model.return_value = user_model

        result = auth_flow.send_forgot_password_code("missing@example.com")

        self.assertTrue(result.success)
        self.assertEqual(result.message, "If the email exists, a reset code has been sent.")

    @patch.object(auth_flow, "timezone")
    @patch.object(auth_flow, "send_email", return_value=None)
    @patch.object(auth_flow, "generate_verification_code", return_value="ABC123")
    @patch.object(auth_flow, "_get_user_meta_model")
    @patch.object(auth_flow, "_get_workflow_setting")
    @patch.object(auth_flow, "get_user_model")
    def test_forgot_password_logs_warning_when_email_not_delivered(
        self,
        mock_get_user_model,
        mock_get_setting,
        mock_get_user_meta_model,
        _mock_generate_code,
        _mock_send_email,
        mock_timezone,
    ):
        now = datetime(2026, 3, 14, 12, 0, tzinfo=timezone.utc)
        mock_timezone.now.return_value = now
        mock_get_setting.side_effect = self._settings_getter(
            ENABLE_EMAIL_DELIVERY=True,
            COMPANY_NAME="Reusable Inc",
            RESET_CODE_TTL_MINUTES=15,
        )

        user_model = self._make_user_model()
        user = SimpleNamespace(email="new@example.com", first_name="New", last_name="User")
        user_model.objects.get.return_value = user
        mock_get_user_model.return_value = user_model

        meta = SimpleNamespace(verify_code=None, verify_time=None, attempts=0, save=MagicMock())
        mock_get_user_meta_model.return_value = self._make_meta_model(meta)

        with patch.object(auth_flow.logger, "warning") as mock_warning:
            result = auth_flow.send_forgot_password_code("new@example.com")

        self.assertTrue(result.success)
        self.assertEqual(result.message, "If the email exists, a reset code has been sent.")
        mock_warning.assert_called_once()

    @patch.object(auth_flow, "timezone")
    @patch.object(auth_flow, "_log_code_for_debug")
    @patch.object(auth_flow, "generate_verification_code", return_value="ABC123")
    @patch.object(auth_flow, "_get_user_meta_model")
    @patch.object(auth_flow, "_get_workflow_setting")
    @patch.object(auth_flow, "get_user_model")
    def test_register_user_disabled_email_delivery_does_not_log_raw_code(
        self,
        mock_get_user_model,
        mock_get_setting,
        mock_get_user_meta_model,
        _mock_generate_code,
        mock_log_code_for_debug,
        mock_timezone,
    ):
        now = datetime(2026, 3, 14, 12, 0, tzinfo=timezone.utc)
        mock_timezone.now.return_value = now
        mock_get_setting.side_effect = self._settings_getter(
            ENABLE_EMAIL_DELIVERY=False,
            COMPANY_NAME="Reusable Inc",
            RESET_CODE_TTL_MINUTES=15,
        )

        user_model = self._make_user_model()
        user_model.objects.filter.return_value.exists.return_value = False
        user = SimpleNamespace(email="new@example.com", first_name="New", last_name="User")
        user_model.objects.create_user.return_value = user
        mock_get_user_model.return_value = user_model

        meta = SimpleNamespace(verify_code=None, verify_time=None, attempts=0, save=MagicMock())
        mock_get_user_meta_model.return_value = self._make_meta_model(meta)

        with patch.object(auth_flow.logger, "info") as mock_info:
            result = auth_flow.register_user_and_send_verification_email(
                email="new@example.com",
                first_name="New",
                last_name="User",
                password="pw",
            )

        self.assertTrue(result.success)
        logged_text = "\n".join(" ".join(map(str, call.args)) for call in mock_info.call_args_list)
        self.assertNotIn("ABC123", logged_text)
        mock_log_code_for_debug.assert_called_once_with(label="Verification", code="ABC123")

    @patch.object(auth_flow, "timezone")
    @patch.object(auth_flow, "_log_code_for_debug")
    @patch.object(auth_flow, "generate_verification_code", return_value="ABC123")
    @patch.object(auth_flow, "_get_user_meta_model")
    @patch.object(auth_flow, "_get_workflow_setting")
    @patch.object(auth_flow, "get_user_model")
    def test_forgot_password_disabled_email_delivery_does_not_log_raw_code(
        self,
        mock_get_user_model,
        mock_get_setting,
        mock_get_user_meta_model,
        _mock_generate_code,
        mock_log_code_for_debug,
        mock_timezone,
    ):
        now = datetime(2026, 3, 14, 12, 0, tzinfo=timezone.utc)
        mock_timezone.now.return_value = now
        mock_get_setting.side_effect = self._settings_getter(
            ENABLE_EMAIL_DELIVERY=False,
            COMPANY_NAME="Reusable Inc",
            RESET_CODE_TTL_MINUTES=15,
        )

        user_model = self._make_user_model()
        user = SimpleNamespace(email="new@example.com", first_name="New", last_name="User")
        user_model.objects.get.return_value = user
        mock_get_user_model.return_value = user_model

        meta = SimpleNamespace(verify_code=None, verify_time=None, attempts=0, save=MagicMock())
        mock_get_user_meta_model.return_value = self._make_meta_model(meta)

        with patch.object(auth_flow.logger, "info") as mock_info:
            result = auth_flow.send_forgot_password_code("new@example.com")

        self.assertTrue(result.success)
        logged_text = "\n".join(" ".join(map(str, call.args)) for call in mock_info.call_args_list)
        self.assertNotIn("ABC123", logged_text)
        mock_log_code_for_debug.assert_called_once_with(label="Password reset", code="ABC123")

    @patch.object(auth_flow, "_get_user_meta_model")
    @patch.object(auth_flow, "get_user_model")
    def test_change_password_updates_user_and_meta(self, mock_get_user_model, mock_get_user_meta_model):
        user_model = self._make_user_model()
        user = SimpleNamespace(set_password=MagicMock(), save=MagicMock())
        user_model.objects.get.return_value = user
        mock_get_user_model.return_value = user_model

        meta = SimpleNamespace(force_password_reset=True, save=MagicMock())
        mock_get_user_meta_model.return_value = self._make_meta_model(meta)

        result = auth_flow.change_password(
            email="person@example.com",
            password="new-secret",
            password_verify="new-secret",
        )

        self.assertTrue(result.success)
        user.set_password.assert_called_once_with("new-secret")
        user.save.assert_called_once()
        self.assertFalse(meta.force_password_reset)
        meta.save.assert_called_once()

    @patch.object(auth_flow, "timezone")
    @patch.object(auth_flow, "_get_user_meta_model")
    @patch.object(auth_flow, "_get_workflow_setting")
    @patch.object(auth_flow, "get_user_model")
    def test_validate_reset_code_wrong_code_increments_attempts(
        self,
        mock_get_user_model,
        mock_get_setting,
        mock_get_user_meta_model,
        mock_timezone,
    ):
        now = datetime(2026, 3, 14, 12, 0, tzinfo=timezone.utc)
        mock_timezone.now.return_value = now
        mock_get_setting.side_effect = self._settings_getter(MAX_VERIFY_ATTEMPTS=3, RESET_CODE_TTL_MINUTES=10)

        user_model = self._make_user_model()
        user_model.objects.get.return_value = SimpleNamespace(is_active=False, save=MagicMock())
        mock_get_user_model.return_value = user_model

        meta = SimpleNamespace(
            verify_code="ABC123",
            verify_time=now - timedelta(minutes=1),
            attempts=0,
            save=MagicMock(),
        )
        mock_get_user_meta_model.return_value = self._make_meta_model(meta)

        result = auth_flow.validate_reset_code(email="person@example.com", code="WRONG")

        self.assertEqual(result.result, auth_flow.VerifyCodeResult.INVALID)
        self.assertEqual(result.attempts_remaining, 2)
        self.assertEqual(meta.attempts, 1)
        meta.save.assert_called_once()

    @patch.object(auth_flow, "timezone")
    @patch.object(auth_flow, "_get_user_meta_model")
    @patch.object(auth_flow, "_get_workflow_setting")
    @patch.object(auth_flow, "get_user_model")
    def test_validate_reset_code_success_activates_user(
        self,
        mock_get_user_model,
        mock_get_setting,
        mock_get_user_meta_model,
        mock_timezone,
    ):
        now = datetime(2026, 3, 14, 12, 0, tzinfo=timezone.utc)
        mock_timezone.now.return_value = now
        mock_get_setting.side_effect = self._settings_getter(MAX_VERIFY_ATTEMPTS=3, RESET_CODE_TTL_MINUTES=10)

        user = SimpleNamespace(is_active=False, save=MagicMock())
        user_model = self._make_user_model()
        user_model.objects.get.return_value = user
        mock_get_user_model.return_value = user_model

        meta = SimpleNamespace(
            verify_code="ABC123",
            verify_time=now - timedelta(minutes=1),
            attempts=0,
            save=MagicMock(),
        )
        mock_get_user_meta_model.return_value = self._make_meta_model(meta)

        result = auth_flow.validate_reset_code(email="person@example.com", code="ABC123")

        self.assertEqual(result.result, auth_flow.VerifyCodeResult.SUCCESS)
        self.assertTrue(user.is_active)
        user.save.assert_called_once()
        self.assertIsNone(meta.verify_code)
        self.assertIsNone(meta.verify_time)
        self.assertEqual(meta.attempts, 0)

    @patch.object(auth_flow, "timezone")
    @patch.object(auth_flow, "_get_user_meta_model")
    @patch.object(auth_flow, "_get_workflow_setting")
    @patch.object(auth_flow, "get_user_model")
    def test_validate_reset_code_hits_too_many_attempts(
        self,
        mock_get_user_model,
        mock_get_setting,
        mock_get_user_meta_model,
        mock_timezone,
    ):
        now = datetime(2026, 3, 14, 12, 0, tzinfo=timezone.utc)
        mock_timezone.now.return_value = now
        mock_get_setting.side_effect = self._settings_getter(MAX_VERIFY_ATTEMPTS=3, RESET_CODE_TTL_MINUTES=10)

        user_model = self._make_user_model()
        user_model.objects.get.return_value = SimpleNamespace(is_active=False, save=MagicMock())
        mock_get_user_model.return_value = user_model

        meta = SimpleNamespace(
            verify_code="ABC123",
            verify_time=now - timedelta(minutes=1),
            attempts=3,
            save=MagicMock(),
        )
        mock_get_user_meta_model.return_value = self._make_meta_model(meta)

        result = auth_flow.validate_reset_code(email="person@example.com", code="ABC123")

        self.assertEqual(result.result, auth_flow.VerifyCodeResult.TOO_MANY_ATTEMPTS)
        self.assertIsNone(meta.verify_code)
        self.assertIsNone(meta.verify_time)
        self.assertEqual(meta.attempts, 0)

    @patch.object(auth_flow, "validate_reset_code")
    def test_verify_reset_code_maps_success_result(self, mock_validate_reset_code):
        mock_validate_reset_code.return_value = auth_flow.VerifyCodeOutcome(
            result=auth_flow.VerifyCodeResult.SUCCESS,
        )

        response = auth_flow.verify_reset_code(email="person@example.com", code="ab12")

        self.assertEqual(response.status_code, int(HTTPStatus.OK))
        self.assertEqual(response.payload, {"detail": "Code verified."})
        mock_validate_reset_code.assert_called_once_with(email="person@example.com", code="AB12")

    @patch.object(auth_flow, "validate_reset_code")
    def test_verify_reset_code_maps_invalid_result(self, mock_validate_reset_code):
        mock_validate_reset_code.return_value = auth_flow.VerifyCodeOutcome(
            result=auth_flow.VerifyCodeResult.INVALID,
            attempts_remaining=1,
        )

        response = auth_flow.verify_reset_code(email="person@example.com", code="ab12")

        self.assertEqual(response.status_code, int(HTTPStatus.ACCEPTED))
        self.assertEqual(response.payload["attempts_remaining"], 1)
        mock_validate_reset_code.assert_called_once_with(email="person@example.com", code="AB12")

    @patch.object(auth_flow, "validate_reset_code")
    def test_verify_reset_code_maps_expired_result(self, mock_validate_reset_code):
        mock_validate_reset_code.return_value = auth_flow.VerifyCodeOutcome(
            result=auth_flow.VerifyCodeResult.EXPIRED,
        )

        response = auth_flow.verify_reset_code(email="person@example.com", code="ab12")

        self.assertEqual(response.status_code, int(HTTPStatus.BAD_REQUEST))
        self.assertEqual(response.payload, {"detail": "Code expired. Please request a new code."})
        mock_validate_reset_code.assert_called_once_with(email="person@example.com", code="AB12")

    @patch.object(auth_flow, "validate_reset_code")
    def test_verify_reset_code_maps_too_many_attempts_result(self, mock_validate_reset_code):
        mock_validate_reset_code.return_value = auth_flow.VerifyCodeOutcome(
            result=auth_flow.VerifyCodeResult.TOO_MANY_ATTEMPTS,
        )

        response = auth_flow.verify_reset_code(email="person@example.com", code="ab12")

        self.assertEqual(response.status_code, int(HTTPStatus.BAD_REQUEST))
        self.assertEqual(response.payload, {"detail": "Too many attempts. Please request a new code."})
        mock_validate_reset_code.assert_called_once_with(email="person@example.com", code="AB12")



import logging
import secrets
import string
from contextlib import nullcontext
from dataclasses import dataclass
from datetime import timedelta
from enum import Enum
from http import HTTPStatus
from typing import Any, Callable, Type

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ImproperlyConfigured
from django.db import IntegrityError, transaction
from django.utils import timezone
from django.utils.module_loading import import_string

from django_workflows.services.email import send_email

logger = logging.getLogger(__name__)

DEFAULT_USER_META_MODEL = "users.models_user_meta.UserMeta"
_AMBIGUOUS_CODE_CHARS = set("IO01")
_ALLOWED_CODE_CHARS = [
    c for c in (string.ascii_uppercase + string.digits) if c not in _AMBIGUOUS_CODE_CHARS
]
_LOG_DEV_CODES_SETTING = "LOG_VERIFICATION_CODES_IN_DEBUG"
_EMAIL_DELIVERY_ENABLED_SETTING = "ENABLE_EMAIL_DELIVERY"


def generate_verification_code(*, length: int = 6) -> str:
    return "".join(secrets.choice(_ALLOWED_CODE_CHARS) for _ in range(length))


def _redact_code(code: str) -> str:
    if not code:
        return ""
    if len(code) <= 2:
        return "*" * len(code)
    if len(code) <= 4:
        return f"{code[0]}{'*' * (len(code) - 2)}{code[-1]}"
    return f"{code[:2]}{'*' * (len(code) - 4)}{code[-2:]}"


def _log_code_for_debug(*, label: str, code: str) -> None:
    # Explicit opt-in for local development only.
    should_log = _is_enabled(_get_workflow_setting(_LOG_DEV_CODES_SETTING, False))
    if should_log and logger.isEnabledFor(logging.DEBUG):
        logger.debug("%s code (masked): %s", label, _redact_code(code))


def _is_delivery_success(response: Any) -> bool:
    status_code = getattr(response, "status_code", None)
    return isinstance(status_code, int) and 200 <= status_code < 300


def _registration_atomic_context():
    if settings.configured:
        return transaction.atomic()
    return nullcontext()


def _register_on_commit(callback: Callable[[], None]) -> bool:
    if not settings.configured:
        return False

    try:
        connection = transaction.get_connection()
    except Exception:
        return False

    if not connection.in_atomic_block:
        return False

    transaction.on_commit(callback)
    return True


def _send_verification_email_or_compensate(*, user: Any, email: str, html: str, enabled: bool) -> bool:
    response = send_email(
        to=[user.email],
        subject="Verify your account",
        html=html,
        enabled=enabled,
    )

    if _is_delivery_success(response):
        logger.info("Verification email sent to %s", user.email)
        return True

    logger.warning(
        "Verification email could not be delivered to %s (response_code=%s)",
        user.email,
        getattr(response, "status_code", None),
    )

    # Compensating action: remove the unverified account so registration can be retried.
    try:
        user.delete()
    except Exception:
        logger.exception(
            "Failed to delete user %s after verification delivery failure",
            email,
        )

    return False


def _workflows_settings() -> dict[str, Any]:
    if not settings.configured:
        return {}

    config = getattr(settings, "WORKFLOWS", {}) or {}
    if not isinstance(config, dict):
        logger.warning("settings.WORKFLOWS must be a dict, got %s", type(config).__name__)
        return {}

    return config


def _get_workflow_setting(name: str, default: Any = None) -> Any:
    return _workflows_settings().get(name, default)


def _is_enabled(value: Any) -> bool:
    if isinstance(value, str):
        return value.strip().lower() == "true"
    return bool(value)


def _as_int(value: Any, default: int) -> int:
    try:
        return int(value)
    except (TypeError, ValueError):
        return default


def _get_user_meta_model() -> Type[Any]:
    model_path = _get_workflow_setting("USER_META_MODEL", DEFAULT_USER_META_MODEL)
    if not isinstance(model_path, str) or "." not in model_path:
        raise ImproperlyConfigured(
            "WORKFLOWS['USER_META_MODEL'] must be a dotted path like 'myapp.models.UserMeta'."
        )

    try:
        return import_string(model_path)
    except Exception as exc:
        raise ImproperlyConfigured(
            f"Unable to import WORKFLOWS['USER_META_MODEL']: {model_path!r}"
        ) from exc

class VerifyCodeResult(Enum):
    SUCCESS = "success"
    INVALID = "invalid"
    EXPIRED = "expired"
    TOO_MANY_ATTEMPTS = "too_many_attempts"

@dataclass(frozen=True)
class ServiceResult:
    success: bool
    message: str | None = None

@dataclass
class VerifyCodeOutcome:
    result: VerifyCodeResult
    attempts_remaining: int | None = None


@dataclass(frozen=True)
class VerifyCodeHttpResult:
    status_code: int
    payload: dict

@dataclass(frozen=True)
class ChangePasswordResult:
    success: bool
    message: str


def register_user_and_send_verification_email(email: str, first_name: str, last_name: str, password: str) -> ServiceResult:
    """
    Registers a new user and sends a verification email.
    """

    User = get_user_model()

    company_name = _get_workflow_setting("COMPANY_NAME", "")
    ttl_minutes = _as_int(_get_workflow_setting("RESET_CODE_TTL_MINUTES", 10), 10)
    enable_email_delivery = _is_enabled(
        _get_workflow_setting(_EMAIL_DELIVERY_ENABLED_SETTING, False)
    )

    if User.objects.filter(email=email).exists():
        return ServiceResult(
            success=False,
            message="A user with this email already exists.",
        )

    user = None
    verification_code = ""
    message = ""

    try:
        with _registration_atomic_context():
            user = User.objects.create_user(
                email=email,
                username=email,
                first_name=first_name,
                last_name=last_name,
                password=password,
                is_active=False,  # User must verify email before activating
            )

            verification_code = generate_verification_code(length=6)
            UserMeta = _get_user_meta_model()

            meta, _ = UserMeta.objects.get_or_create(user=user)
            meta.verify_code = verification_code
            meta.verify_time = timezone.now()
            meta.save()

            message = (
                f"Hello {first_name} {last_name},<br><br>"
                f"Your account verification code is: <strong>{verification_code}</strong><br /><br />"
                f"The code is valid for {ttl_minutes} minutes.<br><br />"
                f"Thank you!<br><br />"
                f"{company_name}"
            )
    except IntegrityError:
        logger.exception("Registration failed due to IntegrityError for email=%s", email)
        return ServiceResult(
            success=False,
            message="Email already exists, login or reset password.",
        )
    except Exception:
        logger.exception("Unexpected error during registration for email=%s", email)
        return ServiceResult(
            success=False,
            message="An unexpected error occurred during registration.",
        )

    if enable_email_delivery:
        def _send_after_commit() -> None:
            _send_verification_email_or_compensate(
                user=user,
                email=email,
                html=message,
                enabled=enable_email_delivery,
            )

        if _register_on_commit(_send_after_commit):
            logger.info(
                "Verification email send scheduled after transaction commit for %s",
                user.email,
            )
        else:
            if not _send_verification_email_or_compensate(
                user=user,
                email=email,
                html=message,
                enabled=enable_email_delivery,
            ):
                return ServiceResult(
                    success=False,
                    message="We could not send a verification email. Please try again later.",
                )
    else:
        logger.info("Email delivery is disabled. Verification email not sent.")
        _log_code_for_debug(label="Verification", code=verification_code)

    return ServiceResult(
        success=True,
        message="Registration successful. Please check your email for the verification code.",
    )

def send_forgot_password_code(email: str) -> ServiceResult:
    """
    Generates and sends a password reset code.
    Always succeeds silently (even if user does not exist).
    """

    User = get_user_model()

    company_name = _get_workflow_setting("COMPANY_NAME", "")
    ttl_minutes = _as_int(_get_workflow_setting("RESET_CODE_TTL_MINUTES", 10), 10)
    enable_email_delivery = _is_enabled(
        _get_workflow_setting(_EMAIL_DELIVERY_ENABLED_SETTING, False)
    )

    try:
        user = User.objects.get(email=email)
    except User.DoesNotExist:
        logger.info(
            "Forgot password requested for non-existent email: %s",
            email,
        )
        return ServiceResult(
            success=True,
            message="If the email exists, a reset code has been sent.",
        )

    UserMeta = _get_user_meta_model()
    meta, _ = UserMeta.objects.get_or_create(user=user)

    code = generate_verification_code(length=6)

    meta.verify_code = code
    meta.verify_time = timezone.now()
    meta.attempts = 0
    meta.save()

    message = (
        f"Hello {user.first_name} {user.last_name},<br><br>"
        f"Your password reset code is: <strong>{code}</strong><br /><br />"
        f"The code is valid for {ttl_minutes} minutes.<br><br>"
        f"Thank you!<br><br>"
        f"{company_name}"
    )

    if enable_email_delivery:
        response = send_email(
            to=[user.email],
            subject="Reset your password",
            html=message,
            enabled=enable_email_delivery,
        )
        if _is_delivery_success(response):
            logger.info("Password reset code sent to %s", user.email)
        else:
            logger.warning(
                "Password reset email could not be delivered to %s (response_code=%s)",
                user.email,
                getattr(response, "status_code", None),
            )
    else:
        logger.info("Email delivery is disabled. Password reset email not sent.")
        _log_code_for_debug(label="Password reset", code=code)

    return ServiceResult(
        success=True,
        message="If the email exists, a reset code has been sent.",
    )

def change_password(*, email: str, password: str, password_verify: str) -> ChangePasswordResult:
    if not email or not password:
        return ChangePasswordResult(False, "email and password are required.")

    if password != password_verify:
        return ChangePasswordResult(False, "Password and password verification do not match.")

    User = get_user_model()
    try:
        user = User.objects.get(email=email)
    except User.DoesNotExist:
        return ChangePasswordResult(False, "Invalid user.")

    user.set_password(password)
    user.save()

    UserMeta = _get_user_meta_model()
    meta, _ = UserMeta.objects.get_or_create(user=user)
    if hasattr(meta, "force_password_reset"):
        meta.force_password_reset = False
    meta.save()

    logger.info(f"Password changed for {email}")
    return ChangePasswordResult(True, "Password changed successfully.")


def verify_reset_code(*, email: str, code: str) -> VerifyCodeHttpResult:
    """API/controller-friendly wrapper around `validate_reset_code`.

    Returns a payload suitable for DRF `Response(...)` and an integer HTTP status code.
    """

    normalized_code = (code or "").upper()
    outcome = validate_reset_code(email=email, code=normalized_code)

    if outcome.result == VerifyCodeResult.SUCCESS:
        return VerifyCodeHttpResult(
            status_code=int(HTTPStatus.OK),
            payload={"detail": "Code verified."},
        )

    if outcome.result == VerifyCodeResult.INVALID:
        payload: dict = {"detail": "Invalid code. Please check your code and try again."}
        if outcome.attempts_remaining is not None:
            payload["attempts_remaining"] = outcome.attempts_remaining

        return VerifyCodeHttpResult(
            status_code=int(HTTPStatus.ACCEPTED),
            payload=payload,
        )

    if outcome.result == VerifyCodeResult.EXPIRED:
        return VerifyCodeHttpResult(
            status_code=int(HTTPStatus.BAD_REQUEST),
            payload={"detail": "Code expired. Please request a new code."},
        )

    if outcome.result == VerifyCodeResult.TOO_MANY_ATTEMPTS:
        return VerifyCodeHttpResult(
            status_code=int(HTTPStatus.BAD_REQUEST),
            payload={"detail": "Too many attempts. Please request a new code."},
        )

    logger.warning("Unhandled VerifyCodeResult: %s", outcome.result)
    return VerifyCodeHttpResult(
        status_code=int(HTTPStatus.BAD_REQUEST),
        payload={"detail": "Unable to verify code."},
    )

def validate_reset_code(email: str, code: str) -> VerifyCodeOutcome:
    User = get_user_model()
    ttl_minutes = _as_int(_get_workflow_setting("RESET_CODE_TTL_MINUTES", 10), 10)
    max_attempts = _as_int(_get_workflow_setting("MAX_VERIFY_ATTEMPTS", 3), 3)

    try:
        user = User.objects.get(email=email)
    except User.DoesNotExist:
        return VerifyCodeOutcome(result=VerifyCodeResult.INVALID)

    UserMeta = _get_user_meta_model()
    meta, _ = UserMeta.objects.get_or_create(user=user)

    if not meta.verify_time or timezone.now() > meta.verify_time + timedelta(minutes=ttl_minutes):
        _clear_meta(meta)
        return VerifyCodeOutcome(result=VerifyCodeResult.EXPIRED)

    if not meta.verify_code:
        _clear_meta(meta)
        return VerifyCodeOutcome(result=VerifyCodeResult.INVALID)

    attempts = int(getattr(meta, "attempts", 0) or 0)

    if attempts >= max_attempts:
        _clear_meta(meta)
        return VerifyCodeOutcome(result=VerifyCodeResult.TOO_MANY_ATTEMPTS)

    if meta.verify_code != code:
        attempts += 1
        meta.attempts = attempts
        meta.save()

        if attempts >= max_attempts:
            _clear_meta(meta)
            return VerifyCodeOutcome(result=VerifyCodeResult.TOO_MANY_ATTEMPTS)

        return VerifyCodeOutcome(
            result=VerifyCodeResult.INVALID,
            attempts_remaining=max_attempts - attempts,
        )

    # success
    _clear_meta(meta)

    # Activate user upon successful verification
    user.is_active = True
    user.save()

    return VerifyCodeOutcome(result=VerifyCodeResult.SUCCESS)


def _clear_meta(meta: Any) -> None:
    meta.verify_code = None
    meta.verify_time = None
    meta.attempts = 0
    meta.save()
